#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2019"
__version__		= "0.1.5"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	pyCalculateMutationFrequencies.py
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import os
import sys
import tempfile
import shutil
import numpy as np
from optparse import *
from collections import defaultdict
from pyCRAC.Classes import NGSFormatWriters
from pyCRAC.Parsers.ParseAlignments import ParseCountersOutput
from pyCRAC.Methods import splitData,processChromFile

def processReadData(datafile,chromosomelength,tempdir="./"):
	""" stores the read data in numpy arrays. """
	data = ParseCountersOutput(datafile)

	chromarray = defaultdict(list)	
	chromarray["+"] = np.zeros((3,chromosomelength),dtype=float)	# one for hits, one for deletions and one for substitutions
	chromarray["-"] = np.zeros((3,chromosomelength),dtype=float)

	while data.readLineByLine():
		chromarray[data.strand][0][data.read_start:data.read_end] += float(data.score)
		if data.substitutions:							# only present in GTF files
			chromarray[data.strand][1][data.substitutions] += float(data.score)
		if data.deletions:
			chromarray[data.strand][2][data.deletions] += float(data.score)
	return chromarray

def mutsAnnotationString(chromarray,strand,start,end,min_frequency=0):
	"""Makes a CIGAR-type string for mutations in reads or clusters"""
	subsstring = str()
	delsstring = str()
	mutlist	   = list()
	subsfreqarray,delsfreqarray = mutsCountsToFrequencies(chromarray,strand,start,end,min_frequency=float(min_frequency))
	for nr in range(len(subsfreqarray)):
		if subsfreqarray[nr]:
			subsstring = "S%.1f" % subsfreqarray[nr]
		if delsfreqarray[nr]:
			delsstring = "D%.1f" % delsfreqarray[nr]
		mutsstring = "%s%s%s" % (start+nr+1,subsstring,delsstring) 
		if subsstring or delsstring:
			mutlist.append(mutsstring)
			subsstring = str()
			delsstring = str()
	if mutlist:
		return ','.join(mutlist)
	else:
		return None

def mutsCountsToFrequencies(chromarray,strand,start,end,min_frequency=0.0):
	""" uses the cluster data to convert mutation numbers to mutation frequencies for each position """
	min_frequency = float(min_frequency)		# set the minimal mutation frequency
	data = chromarray[strand][0][start:end]		# assuming here that 'end' is a 1-based coordinate
	subs = chromarray[strand][1][start:end]
	dels = chromarray[strand][2][start:end]
	subsfreqs = np.ma.fix_invalid(subs*100.0/data,copy=False,fill_value=0.0)	# make an array of substitution frequencies. Remove any NaNs or Infs
	delsfreqs = np.ma.fix_invalid(dels*100.0/data,copy=False,fill_value=0.0)	# make an array of deletion frequencies. Remove any NaNs or Infs
	substitutions = np.where(subsfreqs >= min_frequency,subsfreqs,0.0)			# return all the positions larger than the minimum selected frequency
	deletions	  = np.where(delsfreqs >= min_frequency,delsfreqs,0.0)
	return substitutions,deletions

def main(): 
	parser = OptionParser(usage="usage: %prog [options] -i intervaldata -r readdata", version="%s" % __version__)
	parser.add_option("-i", "--intervaldatafile", dest="intervaldata",help="provide the path to your GTF interval data file.", metavar="intervals.gtf",default=None)
	parser.add_option("-r", "--readdatafile", dest="readdatafile",help="provide the path to your GTF read data file.", metavar="reads.gtf",default=None)
	parser.add_option("-c", "--chromfile",dest="chromfile",help="Location of the chromosome info file. This file should have two columns: first column is the names of the chromosomes, second column is length of the chromosomes.",metavar="yeast.txt",default=None)
	parser.add_option("-o", "--output_file", dest="outfile",help="provide a name for an output file. By default it writes to the standard output", metavar="intervals_with_muts.gtf",default=None)
	parser.add_option("-v", "--verbose",action="store_true",help="to print status messages to a log file",default=False) 
	parser.add_option("--mutsfreq","--mutationfrequency",dest="mutsfreq",action="store",type="int",metavar="10",help="sets the minimal mutations frequency for an interval that you want to have written to our output file. Default = 0%. Example: if the mutsfrequency is set at 10 and an interval position has a mutated in less than 10% of the reads,then the mutation will not be reported.",default=0)
	parser.add_option("--debug",dest="debug",action="store_true",help=SUPPRESS_HELP,default=False)	
	(options, args) = parser.parse_args() 
	outfile = sys.stdout				### default output is standard output
	status	= sys.stdout				### status messages go to the standard output if an output file name has been provided
	logfile = open("pyCalculateMutationFrequencies_log.txt","w")
	if not options.readdatafile:
		parser.error("you forgot to include your GTF read data file")
	if not options.intervaldata:
		parser.error("you forgot to include your GTF interval data file")
	if options.verbose: status	= logfile
	if not options.outfile:
		status = logfile	# to prevent having status messages and data printed to the standard output.
	else:
		outfile = open(options.outfile,"w")
	
	chromlengthfile = open(options.chromfile,"r").readlines()
	chromdict = processChromFile(chromlengthfile)	
	
	### Making a temporary directory ###
	tempdir = tempfile.mkdtemp(prefix="pyCalculateMutationFrequencies_",suffix="_temp",dir="./")
	###
	
	### It is best to split up the data file in temp files, one for each chromosome ### 
	status.write("Creating temporary files for each chromosome in the read data...\n")
	datafiles = splitData(options.readdatafile,tempdir)

	### Making a gtf writer object ###
	gtfwriter = NGSFormatWriters.NGSFileWriter(options.outfile)
	###
		
	### processing the interval data line by line ###
	currentchromosome = str()
	temp = tempfile.NamedTemporaryFile(prefix='intervals',dir=tempdir,suffix='.tmp', delete=False)
	status.write("Sorting interval data...\n")
	os.system("sort -k1,1 -k4,4n %s > %s" % (options.intervaldata,temp.name))
	intervals = ParseCountersOutput(temp.name)
	status.write("Starting analysis...\n")
	while intervals.readLineByLine():
		if intervals.line.startswith("#"): 
			outfile.write(intervals.line)
		chromosome = intervals.chromosome
		if chromosome != currentchromosome:
			status.write("Processing data for chromosome %s...\n" % chromosome)
			file = datafiles[chromosome]
			chromosomelength = chromdict[chromosome]
			chromarray = processReadData(file.name,chromosomelength,tempdir=tempdir)
			currentchromosome = chromosome
		mutsstring = mutsAnnotationString(chromarray,intervals.strand,intervals.read_start,intervals.read_end,min_frequency=options.mutsfreq)
		if mutsstring: 
			outfile.write("%s # %s\n" % (intervals.line.strip(),mutsstring))
		else:
			outfile.write(intervals.line)
	os.unlink(temp.name)
	status.write("Removing remaining temporary files...\n")
	shutil.rmtree(tempdir)	  

if __name__ == "__main__":
	main()


#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2019"
__version__		= "0.0.5"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	pyFastqSplitter.py
#
#
#	Copyright (c) Sander Granneman 2019
#
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

"""
Splits a fastq or fasta file by looking for the character that separates the two DNA sequences. Useful when you collapsed your data

"""

import sys
import six
import gzip
import argparse

def getfilename(filepath):
	""" Returns the file name without path and extension"""
	if isinstance(filepath,file) or not filepath:
			return "standard_input"
	else:
			return filepath.split("/")[-1].split(".")[0]

def headersplitter(string,filetype="fastq"):
	""" splits the header on the @ character"""
	if filetype == "fastq":
		Fld = string.split("@")[1:]
		assert len(Fld) == 2, "Can not split the %s header correctly! %s @ characters in the header\n" % (string, len(Fld))
		header_f = "@%s" % Fld[0]
		header_r = "@%s" % Fld[1]
		return header_f,header_r
	elif filetype == "fasta":
		Fld = string.split(">")[1:]
		if len(Fld) == 2:
			header_f = ">%s" % Fld[0]
			header_r = ">%s" % Fld[1]
		elif len(Fld) == 1:
			header_f = ">%s/1" % Fld[0]
			header_r = ">%s/2" % Fld[0]
		return header_f,header_r
	else:
		raise TypeError("I do not recognize the file type %s. Please use 'fasta' or 'fastq'\n")

def splitFastqFile(inputfile,outputfiles,filetype="fasta",character="|",compressed=False):
	""" Splits fastq files by looking for a specific character that splits the two DNA sequences.
	If this character is not present then it will simply split the DNA sequence in half, provided the length is an even number"""

	### processing the input files ###

	supportedfiletypes = ["fastq","fastq.gz"]

	if not inputfile:										# if the input comes from the standard input
		input = sys.stdin
	elif inputfile and filetype == supportedfiletypes[0]:	# if the input file is not compressed
		input = open(inputfile,"r")
	elif inputfile and filetype == supportedfiletypes[1]:	# if the input file was compressed using gunzip
		input = gzip.open(inputfile,"rb")
	else:
		raise RuntimeError("\n\nUnrecognizable file type %s\n" % filetype)

	### preparing the output files ###

	if len(outputfiles) == 2:
		forward,reverse = outputfiles
		if compressed:
			output_f = gzip.open("%s" % forward, "wb")
			output_r = gzip.open("%s" % reverse, "wb")
		else:
			output_f = open("%s" % forward, "w")
			output_r = open("%s" % reverse, "w")
	else:
		if len(outputfiles) == 1: 	## Only one file name provided. To maintain compatibility with existing pipelines.
			file_name = outputfiles[0]
			forward = "%s_1" % (file_name)
			reverse = "%s_2" % (file_name)
		elif not outputfiles: 			## No file name provided
			file_name = getfilename(inputfile)
			forward = "%s_1" % (file_name)
			reverse = "%s_2" % (file_name,)
		elif not inputfile and not outputfiles:
			forward = "standard_output_1"
			reverse = "standard_output_2"
		else:
			sys.stderr.write("cannot determine what the output files should be called!\n")

		if compressed:
			outtype	 = "fastq.gz"
			output_f = gzip.open("%s.%s" % (forward,outtype), "wb")
			output_r = gzip.open("%s.%s" % (reverse,outtype), "wb")
		else:
			outtype	 = "fastq"
			output_f = open("%s.%s" % (forward,outtype), "w")
			output_r = open("%s.%s" % (reverse,outtype), "w")

	files 	= list()
	files.extend([input,output_f,output_r])			# this list keeps track of all the files/streams

	### iterating over the input files and merging the records ###

	while True:
		try:
			line_f = str()
			line_r = str()
			name	 = str(six.next(input).strip())
			sequence = str(six.next(input).strip())
			six.next(input)
			quality	 = str(six.next(input).strip())

			header_f,header_r = headersplitter(name,filetype="fastq")
			if character:
				seq_f,seq_r	  = sequence.split(character)
				qual_f,qual_r = quality[:len(seq_f)],quality[len(seq_f):]
			else:
				seqlen = len(sequence)
				if seqlen%2!= 0:
					raise RuntimeError("\n\nThe sequence can not be split because the length is an uneven number\n")
				else:
					halfseqlength = len(sequence)/2
					seq_f,seq_r	  = sequence[:halfseqlength],sequence[halfseqlength:]
					qual_f,qual_r = quality[:halfseqlength],quality[halfseqlength:]

			line_f = '%s\n%s\n+\n%s\n' % (header_f,seq_f,qual_f)
			line_r = '%s\n%s\n+\n%s\n' % (header_r,seq_r,qual_r)

			if compressed:
				line_f = line_f.encode()
				line_r = line_r.encode()

			output_f.write(line_f)
			output_r.write(line_r)

		except StopIteration:
			break
	for i in files:
		i.close()

def splitFastaFile(inputfile,outputfiles,filetype="fasta",character="|",compressed=False):
	""" Splits fastq files by looking for a specific character that splits the two DNA sequences.
	If this character is not present then it will simply split the DNA sequence in half, provided the length is an even number"""

	### processing the input files ###

	supportedfiletypes = ["fasta","fasta.gz"]

	if not inputfile:										# if the input comes from the standard input
		input = sys.stdin
	elif inputfile and filetype == supportedfiletypes[0]:	# if the input file is not compressed
		input = open(inputfile,"r")
	elif inputfile and filetype == supportedfiletypes[1]:	# if the input file was compressed using gunzip
		input = gzip.open(inputfile,"rb")
	else:
		raise RuntimeError("\n\nUnrecognizable file type %s\n" % filetype)

	### preparing the output files ###

	if len(outputfiles) == 2:
		forward,reverse = outputfiles
		if compressed:
			output_f = gzip.open("%s" % forward, "wb")
			output_r = gzip.open("%s" % reverse, "wb")
		else:
			output_f = open("%s" % forward, "w")
			output_r = open("%s" % reverse, "w")
	else:
		if len(outputfiles) == 1: 	## Only one file name provided. To maintain compatibility with existing pipelines.
			file_name = outputfiles[0]
			forward = "%s_1" % (file_name)
			reverse = "%s_2" % (file_name)
		elif not outputfiles: 			## No file name provided
			file_name = getfilename(inputfile)
			forward = "%s_1" % (file_name)
			reverse = "%s_2" % (file_name,)
		elif not inputfile and not outputfiles:
			forward = "standard_output_1"
			reverse = "standard_output_2"
		else:
			sys.stderr.write("cannot determine what the output files should be called!\n")

		if compressed:
			outtype	 = "fasta.gz"
			output_f = gzip.open("%s.%s" % (forward,outtype), "wb")
			output_r = gzip.open("%s.%s" % (reverse,outtype), "wb")
		else:
			outtype	 = "fasta"
			output_f = open("%s.%s" % (forward,outtype), "w")
			output_r = open("%s.%s" % (reverse,outtype), "w")

	files	= list()
	files.extend([input,output_f,output_r])			# this list keeps track of all the files/streams

	### iterating over the input files and merging the records ###

	while True:
		try:
			line_f = str()
			line_r = str()
			name	 = str(six.next(input).strip())
			sequence = str(six.next(input).strip())

			header_f,header_r = headersplitter(name,filetype="fasta")
			if character:
				seq_f,seq_r	  = sequence.split(character)
			else:
				seqlen = len(sequence)
				if seqlen%2!=0:
					raise RuntimeError("\n\nThe sequence can not be split becuase the length is an uneven number\n")
				else:
					halfseqlength = len(sequence)/2
					seq_f,seq_r	  = sequence[:halfseqlength],sequence[halfseqlength:]

			line_f = '%s\n%s\n' % (header_f,seq_f)
			line_r = '%s\n%s\n' % (header_r,seq_r)

			if compressed:	# if the output file is compressed:
				line_f = line_f.encode()
				line_r = line_r.encode()

			output_f.write(line_f)
			output_r.write(line_r)

		except StopIteration:
			break

	for i in files:
		i.close()
def main():
	parser = argparse.ArgumentParser(usage="usage: %prog [options] -f file")
	parser.add_argument("-f", "--filename",metavar="fastq_file",dest="filename",help=" To provide the names of two raw data files separated by a single space. Default = standard input",default=None)
	parser.add_argument("-o", "--outfiles",nargs="*",metavar="prefix or forward.fasta,reverse.fasta",dest="outfiles",help="provide either a prefix for the output file names or the full names of the forward and reverse output files.",default=None)
	parser.add_argument("--file_type", dest="filetype",choices=["fasta","fasta.gz","fastq","fastq.gz"],help="type of file, uncompressed or compressed (fastq.gz, gzip/gunzip compressed) fastq. Default is fastq",metavar="FASTQ",default="fastq")
	parser.add_argument("-c", "--character",metavar="|",dest="character",help="If the joined sequences were separated by a specific character then the program can divide the sequences by looking for that character",default=None)
	parser.add_argument("--gz","--gzip", dest="gzip",action="store_true",help="use this option to compress all output file using gunzip or gzip (compression level 9). Note that this slows down the program significantly", default=False)
	args = parser.parse_args()

	supportedfiletypes = ["fasta","fasta.gz","fastq","fastq.gz"]
	if args.filetype in supportedfiletypes[:2]:
		splitFastaFile(args.filename,args.outfiles,args.filetype,args.character,compressed=args.gzip)
	else:
		splitFastqFile(args.filename,args.outfiles,args.filetype,args.character,compressed=args.gzip)

if __name__ == "__main__":
	main()



#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2018"
__version__		= "0.3.5"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@staffmail.ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	SAM.py
#
#
#	Copyright (c) Sander Granneman 2018
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import re
import six
import pprint
from pysam import Samfile
from pyCRAC.Methods import basequality,complement,reverse_complement
from pyCRAC.Classes.Exceptions import *
pp = pprint.PrettyPrinter(indent=4)

MUTS_REGEX = re.compile('[0-9]+[A-Z]|[0-9]+\^[A-Z]+')
#COLLAPSER_REGEX = re.compile('>(\d+)-(\d+)_(.*)')			# group 1 is sequence id and group 2 is number of sequences

def string_lower(string,position):
	a = list(string)
	a[position] = a[position].lower()
	return ''.join(a)
	
def reverse_strand(string):
	if not string:
		return None
	elif string == "F":
		return "R"
	elif string == "R":
		return "F"
	elif string == "+":
		return "-"
	else:
		return "+"

class Process_Sam():
	""" Splits the line from a sam file into columns and adds keys values to a dictionary called interval_dict. Use the type variable to indicate whether the file is a SAM or BAM file"""
	def __init__(self,file_in,base_quality=25):
		""" Initializes the class attributes"""
		self.__file			= None
		self.__base_quality = base_quality
		if hasattr(file_in,'read'):  					# This allows you to pipe data from the standard input. the input file would be sys.stdin
			try:
				self.__file = Samfile("-")
			except ValueError:
				self.__file = Samfile("-","rb")
		else: 
			try:
				self.__file = Samfile(file_in)
			except ValueError:
				self.__file = Samfile(file_in,"rb")
		self.__CIGAR_checked = False
		self.__sam_data		 = self.__file.fetch(until_eof=True)
		self.__line			 = str()
		self.__reset()

	def __reset(self):
		""" Resets all the attributes in the SAM class. Surprisingly,this is much faster than __slots__ or using named tuples! Keeping it simple is usually better!"""
		self.__line			 = str()		# line of the current file
		self.__strand_query	 = str()		# Indicates the strand of the query sequence, default is indicated
		self.__strand_mate	 = str()		# Indicates the strand of the mate sequence, default is indicated
		self.__seqquery		 = str()		# Sequence
		self.__seqmate		 = str()		# Sequence of the mate sequence on the same strand as the reference
		self.__read_type	 = "S"			# Read is either the first read of the pair ("L") or the second/paired ("R"), default is indicated
		self.__qual			 = str()		# Query quality (ASCII-33=Phred base quality)
		self.__alignscore	 = int()		# Align score (Phred-scaled)
		self.__substitutions = list()		# List containing substitutions (0-based positions to reference sequence!)
		self.__insertions	 = list()		# List containing insertions	(0-based positions to reference sequence!)
		self.__deletions	 = list()		# List containing deletions		(0-based positions to reference sequence!)
		self.__skippedref	 = list()		# List containing numbers of nucleotides that were skipped from the reference
		self.__hit_repeat	 = "U"			# Indicates whether a read was mapped only once ("U") or has multiple alignment records ("R"), default is indicated
		self.__rname		 = str()		# Reference sequence name
		self.__num_of_align	 = int()		# Indicates the number of alignment locations for a read on the genome

	def printDictionaryLayout(self):
		""" Prints the interval_dict dictionary layout using pretty printer"""
		pp.pprint(self.__dict__)

	def readline(self,correct_seq=True):
		""" Reads the next line in the file and keeps going until it finds a mapped read """
		self.__reset()
		while True:
			try:
				self.__line = six.next(self.__sam_data)
			except StopIteration:
				return False
			if not self.__line:														# if no line exists: continue
				continue
			if self.__line.seq:														# if no sequence exists: continue
				self.__seqquery = self.__line.seq
			else:
				continue
			if self.__line.tid >= 0:
				self.__rname = self.__file.getrname(self.__line.tid)
			else:	
				continue															# if no reference name exists: continue
			self.__qual		= self.__line.qual
			if self.__line.qual == "." or not self.__line.qual:
				self.__qual = len(self.__line.seq)*"L"								# If there is no quality string, all bases get highest base quality value (Phred score of 41, assuming Illumina 1.8).
			self.decode_CIGAR()
			self.decode_TAGs()
			if self.__line.is_reverse:
				self.__strand_query = "R"
			else: 
				self.__strand_query = "F"
			if self.__line.is_paired:
				if not self.__line.mate_is_reverse:
					self.__strand_mate	= "F"
					self.__strand_query = "R"
				else:
					self.__strand_mate	= "R"
					self.__strand_query = "F"
				if self.__line.is_read1:
					self.__read_type = "L"
				if self.__line.is_read2: 
					self.__read_type = "R"
					
#--------------------------------- Sequences are automatically corrected but this can be turned off! If reads contain deletions, dashes are inserted at the positions and substitutions are highlighted as lowercase letters ---------------------------------------------------
			
			if correct_seq: 
				self.correct_seq()
			return True
									
#-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

	def decode_CIGAR(self):
		""" Decodes the SAM CIGAR info
		Documentation:
		SAM		BAM		Description
		M		0		alignment match (can be a sequence match or mismatch)
		I		1		insertion to the reference
		D		2		deletion from the reference
		N		3		skipped region from the reference
		S		4		soft clipping (clipped sequences present in SEQ)
		H		5		hard clipping (clipped sequences NOT present in SEQ)
		P		6		padding (silent deletion from padded reference)
		=		7		sequence match
		X		8		sequence mismatch
		- H can only be present as the first and/or last operation.
		- S may only have H operations between them and the ends of the CIGAR string.
		- For mRNA-to-genome alignment, an N operation represents an intron. For other types of alignments, the interpretation of N is not defined.
		- Sum of lengths of the M/I/S/=/X operations shall equal the length of SEQ.
		Example:
		CIGAR: 1S29M2D1I25M
		Sequence : TTGCAATTCGCCAGCAAGCACCCAAGGCCTCCGCCAAGTGCACCGTTGCTAGCCT
		position 1 'T', softclip, then 29 matched, then 31 and 32 are deleted, one nucleotide insertion and the remaining 25 are matched
		"""
		start = 0		  # start = 0 is the start position of the read
		self.__CIGAR_checked = True
		cigar_list = self.__line.cigar
		#assert self.__seqquery, "no seq at line:\n %s\n" % self.__line
		sequence = list()
		quality	 = list()
		if type(cigar_list) is list:
			sequence = list(self.__seqquery)
			quality	 = list(self.__qual)
			for (code,position) in cigar_list:
				if	 code == 0:
					start += position
				elif code == 1:
					self.__insertions.extend(list(range(start,start+position)))
					del sequence[start:start+position]
					start += position
				elif code == 3:															   # I have not (yet) come across a read that has this so I haven't been able to test this code properly
					self.__skippedref.append((start,position))		  
				elif code == 4:
					del sequence[start:start+position]
				else:
					continue
			self.__seqquery = "".join(sequence)
			self.__qual		= "".join(quality)
			return True
		else:
			return False
			
	def __call__(self):
		""" Returns the input line as a string"""
		if self.__line:
			return "%s\n" % self.__line
		else:	
			return None

	def decode_TAGs(self):
		"""
		Type	Description
		A		Printable character
		i		Signed 32-bit integer
		f		Single-precision float number
		Z		Printable string
		H		Hex string (high nybble first)

		Tag		Type	Description
		X?		?		Reserved fields for end users (together with Y? and Z?)
		RG		Z		Read group. Value matches the header RG-ID tag if @RG is present in the header.
		LB		Z		Library. Value should be consistent with the header RG-LB tag if @RG is present.
		PU		Z		Platform unit. Value should be consistent with the header RG-PU tag if @RG is present.
		PG		Z		Program that generates the alignment; match the header PG-ID tag if @PG is present.
		AS		i		Alignment score generated by aligner
		SQ		H		Encoded base probabilities for the suboptimal bases at each position (deprecated by E2 and U2)
		OQ		Z		Original base quality. Encoded in the same wasy as QUAL.
		E2		Z		The 2nd most likely base base call.
		U2		Z		Phred-scaled log-lk ratio of the 2nd to the 3rd most likely base calls. Encoding is the same as QUAL.
		MQ		i		The mapping quality score of the mate alignment
		NM		i		Number of nucleotide differences (i.e. edit distance to the reference sequence)
		H0		i		Number of perfect hits
		H1		i		Number of 1-difference hits (an in/del counted as a difference)
		H2		i		Number of 2-difference hits (an in/del counted as a difference)
		UQ		i		Phred likelihood of the read sequence, conditional on the mapping location being correct
		PQ		i		Phred likelihood of the read pair, conditional on both the mapping locations being correct
		NH		i		Number of reported alignments that contains the query in the current record
		IH		i		Number of stored alignments in SAM that contains the query in the current record
		HI		i		Query hit index, indicating the alignment record is the i-th one stored in SAM
		MD		Z		String for mismatching positions in the format of [0-9]+(([ACGTN]|\^[ACGTN]+)[0-9]+)*
		CS		Z		Color read sequence on the same strand as the reference 4
		CQ		Z		Color read quality on the same strand as the reference; encoded in the same way as <QUAL>
		CM		i		Number of color differences 2
		R2		Z		Sequence of the mate.
		Q2		Z		Phred quality for the mate (encoding is the same as <QUAL>).
		S2		H		Encoded base probabilities for the other 3 bases for the mate-pair read. Same encoding as SQ 
		CC		Z		Reference name of the next hit; "=" for the same chromosome
		CP		i		Leftmost coordinate of the next hit
		SM		i		Mapping quality if the read is mapped as a single read rather than as a read pair
		AM		i		Smaller single-end mapping quality of the two reads in a pair
		MF		i		MAQ pair flag (MAQ specific)

		Example:

		AS:i:30 UQ:i:30 NM:i:1	MD:Z:24G9

		Alignment score generated by aligner = 30, Phred likelyhood of the read sequence = 30, Number of nucleotide differences: 1, Mismatching position: 24, mismatched nucleotide: G
		MD example: a string "10A5^AC6" means from the leftmost reference base in the alignment, there are 10 matches followed by an A on the reference which is different from the aligned read base; 
		the next 5 reference bases are matches followed by a 2bp deletion from the reference; the deleted sequence is AC; the last 6 bases are matches. The MD field ought to match the CIGAR string.
		All the information about substitutions and deletions is 0-based!
		"""
		if self.__line.tags:
			for tags in self.__line.tags:
				if tags[0] == "AS"or tags[0] == "UQ":
					self.__alignscore = int(tags[1])
				elif tags[0] == "R2":
					self.__seqmate = tags[1]
				elif tags[0] == "NH":
					if int(tags[1]) == 1:
						self.__hit_repeat == "U"
					else:
						self.__hit_repeat == "R"
					self.__num_of_align = int(tags[1])
				elif tags[0] == "MD":
					start = 0				# start = 0 is the start position of the read
					results = MUTS_REGEX.findall(tags[1])
					if results:
						for i in results:
							subs = re.match("([0-9]+)([A-Z])",i)
							dels = re.match("([0-9]+)\^([A-Z]+)",i)
							if subs:
								substitution = subs.group(2)
								start		+= int(subs.group(1))
								length		 = len(substitution)
								nuc			 = subs.group(2)
								self.__substitutions.append((start,nuc))
								start		+= length
							elif dels:
								start		+= int(dels.group(1))
								deletions	 = dels.group(2)
								length		 = len(deletions)
								self.__deletions.extend([(start+i,deletions[i]) for i in range(length)])
								start		+= length
							else:
								sys.stderr.write("Illegal character in tag search\n")
				else:
					continue
			return True
		else:
			return False
	
	def ismapped(self):
		""" Checks if a read is mapped to the reference sequence """
		if not self.__line.is_unmapped:
			return True
		else:
			return False

	def correct_seq(self):
		""" Corrects the sequence for substitutions and deletions. Insertions and 'softclips' are automatically removed when reading the CIGAR """
		if self.__CIGAR_checked:
			try:
				sequence = list(self.__seqquery)
				quality	 = list(self.__qual)
				for pos in self.__deletions:
					position,deletions = pos[0],pos[1]
					for i in range(len(deletions)):
						sequence.insert(position,"-")
						quality.insert(position,"!")	# ! is base_quality of 0
				for pos in self.__substitutions:
					position,nucleotide = pos[0],pos[1]
					sequence[position]	= sequence[position].lower()
				self.__seqquery = "".join(sequence)
				self.__qual		= "".join(quality)
				return True
			except:
				sys.stderr.write("Error in line %s\tCould not correct the sequence. Please remove from file\n" % self.__line.strip())
				return False
		else:
			sys.stderr.write("you need to run the decode_CIGAR method first before you can use the correct_seq method\n")

	def genomic_seq(self):
		""" Looks if the sequence has any mutations and if so then removes the insertions
		corrects the deletions and substitutions to the genomic sequence and shows these in lowercase."""
		if self.__CIGAR_checked:
			try:
				sequence = list(self.__seqquery)
				for pos in self.__deletions:
					position,deletions = pos[0],pos[1]
					for i in range(len(deletions)):
						sequence[position] = deletions[i].lower()
				for pos in self.__substitutions:
					position,nucleotide = pos[0],pos[1]
					sequence[position]	=  nucleotide.lower()
				return "".join(sequence)
			except:
				sys.stderr.write("Error in line %s\tCould not correct the sequence. Please remove from file\n" % self.__line.strip())
				return sequence
		else:
			sys.stderr.write("you need to run the decode_CIGAR method first before you can use the genomic_seq method\n")

	def deletions(self,chromosome_location=False):
		""" Returns a list of deletions found in a read. Offset of mismatches are 0-based relative 
		to the reference sequence alignment location. If the values are negative then the read is the reverse complement of the reference sequence. 
		If you include the argument 'chromosome_location = True' then the method will return the location of the deletion on the chromosome
		All positions are 0-based positions"""
		if self.__deletions:
			dels = list()
			for i in self.__deletions:
				position,deletion = i[0],i[1]
				dels.extend(list(range(position,position+1)))
			if chromosome_location:
				return [y+self.__line.pos for y in dels]
			else:
				if self.__read_type == "S":
					if self.__strand_query == "F":
						return dels
					if self.__strand_query == "R":
						return [-(y+1) for y in dels]
				elif self.__read_type == "L":
					if self.__strand_query == "F":
						return dels
					if self.__strand_query == "R":
						return [-(y+1) for y in dels]
				elif self.__read_type == "R":
					if self.__strand_query == "F":
						return [-(y+1) for y in dels]
					else:
						return dels
		else:
			return []
					
	def substitutions(self,chromosome_location=False,base_quality=30,filtersubs=None):
		""" Returns a list of substitutions found in a read. Offset of mismatches are 0-based relative 
		to the reference sequence alignment location. If the values are negative then the read is the reverse complement of the reference sequence.\ 
		If you include the argument 'chromosome_location = True' then the method will return the location of the substitutions on the chromosome.\
		If you add basequality=30 then all the substitutions with a basequality lower than 30 will be removed. If you set filtersubs="TC" then\
		only T to C conversions will be kept and others will be removed.\
		All are 0-based positions"""
		if self.__substitutions:
			subs = list()
			for i in self.__substitutions:
				position = i[0]
				subs.append(position)
			if chromosome_location:
				return [y+self.__line.pos for y in subs]
			else:
				if self.__read_type == "S":
					if self.__strand_query == "F":
						return subs
					if self.__strand_query == "R":
						return [-(y+1) for y in subs]
				elif self.__read_type == "L":
					if self.__strand_query == "F":
						return subs
					if self.__strand_query == "R":
						return [-(y+1) for y in subs]
				elif self.__read_type == "R":
					if self.__strand_query == "F":
						return [-(y+1) for y in subs]
					else:
						return subs
		else:
			return []

	def filterMutations(self,filter):
		"""Checks if a read has specific deletions or substitutions:
		Choices are:
				'delsonly': checks if a read has deletions.		 If so, it returns True and only deletions will be reported. Else, it returns False.
				'subsonly': checks if a read has substitutions.	 If so, it returns True and only substitutions will be reported. Else, it returns False.
				'nomuts'  : no mutations will be reported.
				'TC'	  : checks if a read has T->C mutations. If so, it removes mutations that do not fit these criteria and returns True, else False.
				'[T,C,G]' : (list) checks if the reads has T's, C's and/or G's mutated. If so, it removes mutations that do not fit these criteria and returns True, else False.
		"""
		if filter	== "delsonly":
			self.__substitutions = []
			if self.__deletions: 
				return True
			else: 
				return False
		elif filter == "subsonly" and self.__substitutions:
			self.__deletions = []
			if self.__substitutions: 
				return True
			else: 
				return False
		elif filter == "nomuts":
			self.__substitutions = []
			self.__deletions	 = []
			return True

		elif isinstance(filter,list):
			if self.__substitutions:
				newmutslist = list()
				for (position,nucleotide) in self.__substitutions:
					if self.__read_type == "S" or self.__read_type == "L":
						if self.__strand_query == "R":
							nucleotide = reverse_complement(nucleotide)
					elif self.__read_type == "R":
						if self.__strand_query == "F":
							nucleotide = reverse_complement(nucleotide) 
					if nucleotide.upper() in filter:
						newmutslist.append((position,nucleotide))
				if newmutslist: 
					self.__substitutions = newmutslist
					return True
				else:
					return False
						
			if self.__deletions:
				newmutslist = list()
				for (position,nucleotide) in self.__deletions:
					if self.__read_type == "S" or self.__read_type == "L":
						if self.__strand_query == "R":
							nucleotide = reverse_complement(nucleotide)
					elif self.__read_type == "R":
						if self.__strand_query == "F":
							nucleotide = reverse_complement(nucleotide) 
					if nucleotide.upper() in filter:
						newmutslist.append((position,nucleotide))
				if newmutslist: 
					self.__deletions = newmutslist	  
					return True
				else:
					return False	
			
			mutations = self.__substitutions + self.__deletions
			if mutations:
				return True						   
			else:
				return False

		elif isinstance(filter,str) and len(filter) == 2:
			self.__deletions = []
			if self.__substitutions:
				newmutslist = list()
				for subs in self.__substitutions:
					(position,nucleotide) = subs
					mutation = self.sequence()[position]		# read base
					actual	 = nucleotide						# ref base
					if self.__read_type == "S" or self.__read_type == "L":
						if self.__strand_query == "R":
							mutation = self.sequence()[-(position+1)]		# read base
							actual	 = reverse_complement(nucleotide)		# ref base
					elif self.__read_type == "R":
						if self.__strand_query == "F":
							mutation = self.sequence()[-(position+1)]
							actual	 = reverse_complement(nucleotide)
					string = "%s%s" % (actual.upper(),mutation.upper())
					if string == filter:
						newmutslist.append(subs)		# keep only those substitutions that match your criteria
				self.__substitutions = newmutslist
				if newmutslist:
					return True
				else:
					return False
			else:
				return False
		else:
			return False
					
	def subsBaseQualFilter(self,basequality=0):
		""" checks the base quality of substitutions in the substitutions list. Returns a filtered list """
		for subs in sorted(self.__substitutions):
			if self.baseQuality(subs[0]) < basequality:
				self.__substitutions.remove(subs)
		return True

	def baseQuality(self,nucl_nr,qualtype="L"):
		""" This method returns the basequality value (an integer, NOT ASCI!) at position nucl_nr. Usage: base_quality(24)
		You need to include format of the quality scoring, for example: qualtype="sanger" or qualtype="illumina
		S - Sanger		  Phred+33,	 raw reads typically (0, 40)
		X - Solexa		  Solexa+64, raw reads typically (-5, 40)
		I - Illumina 1.3+ Phred+64,	 raw reads typically (0, 40)
		J - Illumina 1.5+ Phred+64,	 raw reads typically (3, 40) with 0=unused, 1=unused, 2=Read Segment Quality Control Indicator (bold) 
		L - Illumina 1.8+ Phred+33,	 raw reads typically (0, 41)
		Default is L """
		quality = self.__qual
		if len(quality) > 1:
			if nucl_nr > len(quality):
				corr_nr = nucl_nr - self.__line.pos
				if corr_nr < 0:
					raise BaseQualityError("\n\ncould not calculate the base-quality for nucleotide position %d in read %s\nQuality length\t%s\nposition\t%s" % (nucl_nr,self.__line.qname,len(quality),nucl_nr))
				else:
					return basequality(quality[corr_nr],ftype=qualtype)
			else:
				return basequality(quality[nucl_nr],ftype=qualtype)
		else:
			return 0

	def baseQualityString(self):
		""" Returns the quality string for the sequence in the line"""
		if self.__read_type == "S" or self.__read_type == "L":
			if self.__strand_query == "F":
				return self.__qual
			else:
				return self.__qual[::-1]
		elif self.__read_type == "R":
			if self.__strand_query == "F":
				return self.__qual[::-1]
			else:
				return self.__qual
			
	def mutations(self):
		""" This method returns a list of chromosomal coordinates for all the mutations found in a read sequence"""
		mutations = self.__substitutions + self.__insertions + self.__deletions
		if mutations:
			return mutations
		else:
			return None
							
	def sequence(self):
		""" Returns the sequence of the reads. The orientation depends on whether single end or paired
		end sequencing was performed and to which strand the read was mapped. Note that in the SAM file all 
		mapped reads are represented on the forward genomic strand. The bases are reverse complemented from the 
		unmapped read sequence and the quality scores and cigar strings are recorded consistently with the bases. """
		if self.__read_type == "S" or self.__read_type == "L":
			if self.__strand_query == "F":
				return self.__seqquery
			else:
				return reverse_complement(self.__seqquery)
		elif self.__read_type == "R":
			if self.__strand_query == "F":
				return reverse_complement(self.__seqquery)
			else:
				return self.__seqquery

	def uniqueAlnLocation(self):
		""" Returns 'True' if a read has only one genomic location. Otherwise returns 'False''"""
		if self.__hit_repeat == "U":
			return True
		else:
			return False
					
	def seq_ID(self):
		""" Returns the sequence identification number of the read"""
		if self.__line.qname.startswith(">"):
			return self.__line.qname[1:]
		else:
			return self.__line.qname

	def read_type(self):
		""" Rreturns the read_type indicator ("S" = single end read; "L" or "R" indicate paired end reads)"""
		return self.__read_type
	
	def align_score(self):
		""" Returns the Phred format alignment score 10log10(P(R|Ai)).
		For status of 'R' and when not report alignment locations for repeats,
		the alignment score becomes the number of alignments to the read.
		For paired end the alignment score includes the fragment length penalty"""
		if self.__alignscore:
			return self.__alignscore
		else:
			return 0

	def alignScoreThreshold(self,number):
		""" Bool method. Checks whether the alignment score is smaller or equal to the to a threshold set by the
		user. If the alignment score is smaller than the threshold, the method returns 'False' else it returns 'True' """
		if number > self.__alignscore:
			return False
		else:
			return True

	def align_quality(self):
		""" MAPping Quality (phred-scaled posterior probability that the mapping position of this read is incorrect)
		Field MAPQ considers pairing in calculation if the read is paired. Providing MAPQ is recommended. If such a
		calculation is difficult, 255 should be applied, indicating the mapping quality is not available."""
		if int(self.__line.mapq):
			return int(self.__line.mapq)
		else:
			return 0

	def alignQualityThreshold(self,number):
		""" Checks whether the alignment quality is smallor or equal to the to a threshold set by the
		user. If the alignment quality is smaller than the threshold, the method returns 'False' else it returns 'True' """
		if number > int(self.__line.mapq):
			return False
		else:
			return True

	def lengthThreshold(self,number):
		""" Tests whether the length of the read sequence is smaller or equal to a maximum length
		threshold set by the user. If the read is smaller than the threshold the method returns 'True', else
		it returns 'False' """
		if number > len(self.__seqquery):
			return True
		else:
			return False

	def unique_hits(self):
		""" Returns R or U. R indicates that the read was mapped to multiple genomic locations.
		U indicates that the read mapped to only one location on the genome"""
		if self.__num_of_align == 1:
			return "U"
		else:
			return "R"
					
	def sequenceAlignmentLocations(self):
		""" Returns the number of alignments to the read"""
		return self.__num_of_align

	def read_start(self):
		""" Returns a left most 0-based start position of the read on the chromosome"""
		if self.__line.pos >= 0:
			return self.__line.pos	# although SAM start coordinates are 1-based, pysam coordinates are ALWAYS 0-based!!!
		else:
			return None

	def read_end(self):
		""" Returns the 0-based end position of the read on the chromosome"""
		if self.__line.pos >= 0:
			return self.__line.pos + len(self.__seqquery)
		else:
			return None

	def read_start_paired(self):
		""" Returns the left most 0-based start position of the paired read, if there is one"""
		if self.__line.mpos >= 0:
			return self.__line.mpos # although SAM start coordinates are 1-based, pysam coordinates are ALWAYS 0-based!!!
		else:
			return None
					
	def read_end_paired(self):
		""" Returns the 0-based end position of the paired read, if there is one"""
		if self.__seqmate:
			if self.__line.mpos <= self.__line.pos:
				return self.__line.mpos + len(self.__seqmate) - 1
			else:
				return None 
									   
	def paired_sequence(self,distance=1000):
		""" This BOOL method returns TRUE if the read has a paired read mapped to the same chromosomal location"""
		if self.__line.is_paired and self.__line.mpos >= 0 and not self.__line.mate_is_unmapped and self.__line.isize > -(distance) and self.__line.isize < distance:
			return True
		else:
			return False
					
	def strand(self):
		""" Returns the strand of the chromosome ("F" = plus, "R" = minus strand) the read was mapped to
		Note that for paired end reads the strand of the reverse sequencing reaction will be reversed"""
		if self.__read_type == "R":
			return self.__strand_mate
		else:
			return self.__strand_query

	def strand_paired(self):
		""" Returns the strand of the paired read if there is one"""
		if self.__line.is_paired and self.__strand_mate:
			return self.__strand_mate
		else:
			return None
	
	def chr_ID(self):
		""" Returns the chromosome the read sequence was mapped to"""
		return self.__rname
	
	def seq_length(self):
		""" Returns the length of the read sequence"""
		return len(self.__seqquery)

	def fasta(self):
		""" Returns the read ID and read sequence in the fasta format"""
		if self.__line.qname:
			fasta =	 ">%s\n%s" % (self.__line.qname,self.__seqquery)
			return fasta
	
	def tab(self):
		""" Returns the read ID and read sequence in the tab delimited format"""
		if self.__line.qname:
			tab =  "%s\t%s" % (self.__line.qname,self.__seqquery)
			return tab


#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2019"
__version__		= "0.0.7"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "beta"

##################################################################################
#
#	pyNormalizeIntervalLengths.py
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import os
import sys
from optparse import *
from pyCRAC.Methods import processChromFile
	
def normalizeIntervalLength(start,end,chromosomelength,length=20):
	""" extends or trims the interval coordinates to a set length. Default = 20 bp. """
	newstart = int()
	newend	 = int()
	if start - length < 0: start = 1							# to make sure that the start is always a positive number 
	if end + length > chromosomelength: end = chromosomelength	# to make sure that interval doesn't go beyond the chromosome boundaries.
	actlength  = end - start
	difference = length - actlength
	if difference == 0:
		return start,end
	else:
		newstart = round(float(start) - float(difference)/2.0)
		if newstart < 0: newstart = 1
		newend	= round(float(end) + float(difference)/2.0)
		if newend > chromosomelength: newend = chromosomelength
		return int(newstart),int(newend)		# convert back to integers
	
def extendIntervalLength(start,end,chromosomelength,length=20):
	""" extends or the interval coordinates to a minimum length. Default = 20 bp. """
	if start - length < 0: start = 1							# to make sure that the start is always a positive number 
	if end + length > chromosomelength: end = chromosomelength	# to make sure that interval doesn't go beyond the chromosome boundaries.
	newstart = int()
	newend	 = int()
	actlength  = end - start
	difference = length - actlength
	if difference > 0:
		newstart = round(float(start) - float(difference)/2.0)
		if newstart < 0: newstart = 1
		newend = round(float(end) + float(difference)/2.0)
		return int(newstart),int(newend)		# convert back to integers
	else:
		return start,end

	return int(newstart),int(newend)			# convert back to integers
	
def addToInterval(start,end,chromosomelength,length=20):
	""" extends each interval to with a specific length. """
	start -= length
	end += length
	if start <= 0: start = 1
	if end > chromosomelength: end = chromosomelength
	return start,end
	
def addToLeft(start,end,strand,chromosomelength,length=20):
	""" adds a specific length to the 5'end of a gene """
	if strand == "+":
		start -= length
		if start <= 0: start = 1
	else:
		end += length
		if end > chromosomelength: end = chromosomelength
	return start,end
	
def addToRight(start,end,strand,chromosomelength,length=20):
	""" adds a specific length to the 5'end of a gene """
	if strand == "-":
		start -= length
		if start <= 0: start = 1
	else:
		end += length
		if end > chromosomelength: end = chromosomelength
	return start,end
		
def main():
	parser = OptionParser(usage="usage: %prog [options] -f intervaldata -c chromfile -o output_file.gtf",version="%s" % __version__)
	parser.add_option("-f", "--intervaldata", dest="intervaldata",help="Name of the bed/gff/gtf file containing the read/cDNA coordinates", metavar="mygtffile",default=None)	 
	parser.add_option("--file_type",dest="file_type",choices=["bed","gtf","gff"],help="this tool supports bed6, gtf and gff input files. Please select from 'bed','gtf' or 'gff'. Default=gtf",default="gtf")	
	parser.add_option("-o", "--outfile",dest="outfile",help="Optional. Provide the name of the output file. Default is 'selected_intervals.gtf'",metavar="outfile.gtf",default=None) 
	parser.add_option("-c", "--chromfile",dest="chromfile",help="Location of the chromosome info file. This file should have two columns: first column is the names of the chromosomes, second column is length of the chromosomes. Default is yeast",metavar="yeast.txt",default=None)
	parser.add_option("-v", "--verbose",action="store_true",help="to print status messages to a log file",default=False)
	parser.add_option("--fixed",type="int",dest="length",help="to set a fixed length for each interval in the gtf file.",metavar=20,default=0)
	parser.add_option("--min",dest="min",type="int",help="to set a minimal length for an interval. If an interval is shorter than the set minimal length it will be extended. Default = OFF.",metavar=20,default=0)
	parser.add_option("--addboth",type="int",dest="add",help="to extend the coordinates on both sides with a fixed number.",metavar=20,default=0)
	parser.add_option("--addleft",type="int",dest="left",help="to extend 5' end with a fixed number",metavar=20,default=0)
	parser.add_option("--addright",type="int",dest="right",help="to extend 3' end with a fixed number",metavar=20,default=0)
	(options, args) = parser.parse_args()
	outfile = sys.stdout
	if options.outfile:
		outfile = open(options.outfile,"w")
	chromlengthfile = open(options.chromfile,"r").readlines()
	chromdict = processChromFile(chromlengthfile)
	data  = open(options.intervaldata,"r").readlines()
	for line in data:
		if line.startswith("#"):
			outfile.write(line)
			continue
		Fld = line.strip().split("\t")
		chromosome = Fld[0]
		try:
			if options.file_type == "bed":
				strand = Fld[5]
				if options.length:
					Fld[1],Fld[2] = normalizeIntervalLength(int(Fld[1]),int(Fld[2]),chromdict[chromosome],options.length)
				elif options.min:
					Fld[1],Fld[2] = extendIntervalLength(int(Fld[1]),int(Fld[2]),chromdict[chromosome],options.min)
				elif options.add:
					Fld[1],Fld[2] = addToInterval(int(Fld[1]),int(Fld[2]),chromdict[chromosome],options.add)
				elif options.left:
					Fld[1],Fld[2] = addToLeft(int(Fld[1]),int(Fld[2]),chromdict[chromosome],options.left)
				elif options.right:
					Fld[1],Fld[2] = addToRight(int(Fld[1]),int(Fld[2]),chromdict[chromosome],options.right)
			else:
				strand = Fld[6]
				if options.length:
					Fld[3],Fld[4] = normalizeIntervalLength(int(Fld[3]),int(Fld[4]),chromdict[chromosome],options.length)
				elif options.min:
					Fld[3],Fld[4] = extendIntervalLength(int(Fld[3]),int(Fld[4]),chromdict[chromosome],options.min)
				elif options.add:
					Fld[3],Fld[4] = addToInterval(int(Fld[3]),int(Fld[4]),chromdict[chromosome],options.add)
				elif options.left:
					Fld[3],Fld[4] = addToLeft(int(Fld[3]),int(Fld[4]),chromdict[chromosome],options.left)
				elif options.right:
					Fld[3],Fld[4] = addToRight(int(Fld[3]),int(Fld[4]),chromdict[chromosome],options.right)
		except:
			sys.stderr.write("Could not process line %s\nPlease check the input file.\n" % line)
			continue
				
		outfile.write("%s\n" % "\t".join([str(i) for i in Fld]))					
	
if __name__ == "__main__":
	main()


#!/usr/bin/python

__author__      = "Sander Granneman"
__copyright__   = "Copyright 2020"
__version__     = "1.5.1"
__credits__     = ["Sander Granneman","Grzegorz Kudla","Hywell Dunn Davies"]
__maintainer__  = ["Sander Granneman"]
__email__       = "sgrannem@staffmail.ed.ac.uk"
__status__      = "Production"


import re
import sys
import os

try:
	from setuptools import setup
	from setuptools.command import easy_install
	sys.stdout.write("Python development and setuptools have been installed...\n")
except:
	sys.stderr.write("Python development and setuptools have not been installed on this machine\nPlease contact the admin of this computer to install these modules\n")
	exit()

setup(name='pyCRAC',
	version='%s' % __version__,
	description='Python NextGen sequencing data processing software',
	author='Sander Granneman',
	author_email='sgrannem@staffmail.ed.ac.uk',
	url='http://sandergranneman.bio.ed.ac.uk/Granneman_Lab/pyCRAC_software.html',
	packages=['pyCRAC','pyCRAC.Parsers','pyCRAC.Classes','pyCRAC.Methods'],
	install_requires=['numpy >= 1.5.1', 'cython >=0.19', 'pysam >= 0.6','six >= 1.9.0'],
	scripts=[
					'pyCRAC/pyReadAligner.py',
					'pyCRAC/pyMotif.py',
					'pyCRAC/pyPileup.py',
					'pyCRAC/pyBarcodeFilter.py',
					'pyCRAC/pyReadCounters.py',
					'pyCRAC/pyBinCollector.py',
					'pyCRAC/pyCalculateFDRs.py',
					'pyCRAC/pyClusterReads.py',
					'pyCRAC/pyCalculateMutationFrequencies.py',
					'pyCRAC/scripts/pyCalculateChromosomeLengths.py',
					'pyCRAC/scripts/pyFastqDuplicateRemover.py',
					'pyCRAC/scripts/pyAlignment2Tab.py',
					'pyCRAC/scripts/pyGetGTFSources.py',
					'pyCRAC/scripts/pySelectMotifsFromGTF.py',
					'pyCRAC/scripts/pyFasta2tab.py',
					'pyCRAC/scripts/pyFastqJoiner.py',
					'pyCRAC/scripts/pyFastqSplitter.py',
					'pyCRAC/scripts/pyExtractLinesFromGTF.py',
					'pyCRAC/scripts/pyGetGeneNamesFromGTF.py',
					'pyCRAC/scripts/pyCheckGTFfile.py',
					'pyCRAC/scripts/pybed2GTF.py',
					'pyCRAC/scripts/pyGTF2sgr.py',
					'pyCRAC/scripts/pyGTF2bed.py',
					'pyCRAC/scripts/pyGTF2bedGraph.py',
					'pyCRAC/scripts/pyFilterGTF.py',
					'pyCRAC/scripts/pyNormalizeIntervalLengths.py',
				],
	classifiers=[   'Development Status :: 5 - Production/Stable',
					'Environment :: Console',
					'Intended Audience :: Education',
					'Intended Audience :: Developers',
					'Intended Audience :: Science/Research',
					'License :: Freeware',
					'Operating System :: MacOS :: MacOS X',
					'Operating System :: POSIX',
					'Programming Language :: Python :: 3.6',
					'Topic :: Scientific/Engineering :: Bio-Informatics',
					'Topic :: Software Development :: Libraries :: Application Frameworks'
				],
	data_files=[    ('pyCRAC/db/',[
					'pyCRAC/db/Saccharomyces_cerevisiae.EF2.59.1.0_chr_lengths.txt',
					'pyCRAC/db/Saccharomyces_cerevisiae.EF2.59.1.3.gtf',
					'pyCRAC/db/Saccharomyces_cerevisiae.EF2.59.1.0.fa',
					'pyCRAC/db/Saccharomyces_cerevisiae.EF2.59.1.0.fa.tab']),
					('pyCRAC/tests/',[
					'tests/test.novo',
					'tests/test.sh',
					'tests/test_coordinates.txt',
					'tests/test.gtf',
					'tests/test_f.fastq',
					'tests/test_f.fastq.gz',
					'tests/test_f_dm.fastq',
					'tests/test_r.fastq',
					'tests/test_r.fastq.gz',
					'tests/test_r_dm.fastq',
					'tests/indexes.txt',
					'tests/barcodes.txt',
					'tests/genes.list']),
					('pyCRAC',[
					'The_pyCRAC_Manual.pdf'])
				]
			  )



#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2020"
__version__		= "0.0.6"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	pyGTF2bed.py
#
#
#	Copyright (c) Sander Granneman 2019
#
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import re
from optparse import *
from pyCRAC.Classes.NGSFormatWriters import NGSFileWriter
from pyCRAC.Classes.NGSFormatReaders import NGSFileReader

def gtf2bed(gtf_file,outfile,name=None,description=None,color="black",colorbystrand="red,blue"):
	""" converts gtf files to bed6 files the way I like it """
	gtfreader = NGSFileReader(gtf_file)		# gtf reader object
	bedwriter = NGSFileWriter(outfile)		# bed writer object
	if name or description:
		if not name: name = "User_Supplied_Track"
		if not description: description = "User_Supplied_Track"
		bedwriter.writeTrackLine("bed",name,description,colorbystrand)	# write trackline information for visualization in the UCSC genome browser
	while gtfreader.readGTFLine():
		names = list()
		if gtfreader.gene_name:
			names.append(gtfreader.gene_name)
		elif gtfreader.gene_id:
			names.append(gtfreader.gene_id)
		if gtfreader.transcript_name:
			names.append(gtfreader.transcript_name)
		elif gtfreader.transcript_id:
			names.append(gtfreader.transcript_id)
		if names:
			names = list(set(names))	# remove any duplicates, for example if the gene_name and transcript_name are identical
		else:
			names = [name]			# if it can't find any attributes, just print the name of the experiment in the output file
		# write to output file:

		bedwriter.writeBed(gtfreader.chromosome,gtfreader.start,gtfreader.end,gtfreader.strand,",".join(names),gtfreader.score)

def main():
	parser = OptionParser(usage="usage: %prog [options] --gtf=gtf_file.gtf", version="%s" % __version__)
	files = OptionGroup(parser, "File input options")
	files.add_option("--gtf",dest="gtf_file",metavar="Yourfavoritegtf.gtf",help="type the path to the gtf file that you want to convert. Default is standard input",default=None)
	files.add_option("-o","--outfile",dest="outfile",help="type the name and path of the file you want to write the output to. Default is standard output",default=None)
	ucsc = OptionGroup(parser, "These options can be used to add and modify a track line for the UCSC genome browser")
	ucsc.add_option("-n","--name",dest="name",help="For the USCS track line: provide a track name. Default is none ",default=None)
	ucsc.add_option("-d","--description",dest="description",help="For the USCS track line: provide a track description. Default is none ",default=None)
	ucsc.add_option("--color",dest="color",help="select the track color. Default = black",default="black")
	ucsc.add_option("-s","--colorstrands",dest="strands",help="select the colors for each strand. Default = 'red,blue'",default="red,blue")
	parser.add_option_group(ucsc)
	parser.add_option_group(files)
	(options, args) = parser.parse_args()
	data = sys.stdin
	out	 = sys.stdout

	if options.gtf_file: data = open(options.gtf_file,"r")
	if options.outfile : out = open(options.outfile,"w")

	gtf2bed(data,out,name=options.name,description=options.description,color=options.color,colorbystrand=options.strands)

if __name__ == "__main__":
	main()



#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2019"
__version__		= "0.0.3"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	pyFasta2Tab.py
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import os
from optparse import *
from pyCRAC.Methods import getfilename

def main():
	""" this program converts fasta files to the tabular output """
	parser = OptionParser(usage="usage: %prog -f filename.fasta", version="%s" % __version__)
	parser.add_option("-f", "--input_file",dest="fasta_file",metavar="fasta_file",help="provide the name and path of your fasta input file. Default is standard input.",default=None)
	parser.add_option("-o", "--output_file",dest="tab_file",metavar="tab_file",help="provide a name and path for your tab output file. Default is input with .tab extension",default=None),
	(options, args) = parser.parse_args()
	input = sys.stdin
	output_file = "%s.tab" % getfilename(input)
	output = None
	if options.fasta_file:
		input = open(options.fasta_file,"r")
		output_file = "%s.tab" % getfilename(options.fasta_file)
	if options.tab_file:
		output = open(options.tab_file,"w")
	else:
		output = open(output_file,"w")
	id = str()
	sequence = str()
	for line in input:
		if line.startswith(">") and sequence:		# if a new id has been found and sequence exists, print to file
			output.write("%s\t%s\n" % (id,sequence))
			id = line[1:].strip()
			sequence = str()
		elif line.startswith(">"):					# if only a new id has been found, store the id
			id = line[1:].strip()
		else:
			sequence += line.strip()				# else, concatenate the sequence strings
	output.write("%s\t%s\n" % (id,sequence))
		
if __name__ == "__main__":
	main()


#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2020"
__version__		= "0.0.5"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	pyGetGeneNamesFromGTF.py
#
#
#	Copyright (c) Sander Granneman 2020
#
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import re
from optparse import *
from collections import defaultdict

def getGenes(line,attribute="gene_name"):
	"""Looks for the gene_name in attributes and returns associated gene names, if any"""
	gene_regex	   = re.compile("\"(.*)\"")
	altgene_regex  = re.compile("=(.*)")
	attributefield = list()
	genes		   = list()
	try:
		attributefield = line.split("\t")[8]
	except IndexError:
		return []
	attribute_list = attributefield.split(";")
	for i in attribute_list:
		if re.search("\\b%s\\b" % attribute,i):
			genes = gene_regex.findall(i)
			if not genes:
				genes = altgene_regex.findall(i)
			if genes:
				return genes[0].split(",")
	if not genes:
		for i in attribute_list:
			if re.search("gene_id",i):
				genes = gene_regex.findall(i)
				if not genes:
					genes = altgene_regex.findall(i)
				if genes:
					return genes[0].split(",")
	else:
		return []

def returnGenes(data,attribute="gene_name",totalcount=False):
	"""Returns a list with gene names"""
	genelist	 = set()
	gene_counter = defaultdict(int)
	for line in data:
		if line[0] != "#":
			genes = getGenes(line,attribute)
			if genes:
				for i in genes:
					if totalcount:
						gene_counter[i] += int(line.strip().split("\t")[5])
					else:
						gene_counter[i] += 1
					genelist.add(i)
			else:
				sys.stdout.write("No genes found in line:\n%s" % line)
	if not genelist:
		sys.stderr.write("could not find any %s attributes.\nPlease choose a different attribute\n" % (attribute))

	return genelist,gene_counter

def main():
	parser = OptionParser(usage="usage: %prog [options] --gtf=myfavgtf --count", version="%s" % __version__)
	parser.add_option("--gtf",dest="gtf_file",metavar="Yourfavoritegtf.gtf",help="type the path to the gtf file that you want to use. By default it expects data from the standard input.",default=None)
	parser.add_option("-o","--outfile",dest="outfile",help="type the name and path of the file you want to write the output to. Default is standard output",default=None)		
	parser.add_option("-a","--attribute", dest="attribute",choices=["gene_name","gene_id","transcript_name","transcript_id"], help="from which attribute do you want to extract names? Choices: gene_name, gene_id, transcript_name, transcript_id", default="gene_name")
	parser.add_option("--count",dest="count",action="store_true",help="To count the occurence for each source/annotation in the gtf file",default=False)
	parser.add_option("--totalcount",dest="total",action="store_true",help="To count the occurence for each source/annotation in the gtf file including the total number of intervals mapped to that feature ",default=False)
	(options, args) = parser.parse_args()
	data = sys.stdin
	out	 = sys.stdout
	getcounts = False
	if options.outfile:
		out = open(options.outfile,"w")
	if options.gtf_file:
		data = open(options.gtf_file,"r")
	if options.count or options.total:
		getcounts = True
	genes,counter = returnGenes(data,options.attribute,totalcount=options.total)
	if genes:
		if getcounts:
			for name in sorted(genes):
				out.write("%s\t%s\n" % (name,counter[name]))
		else:
			for name in sorted(genes):
				out.write("%s\n" % name)

if __name__ == "__main__":
	main()



#!/usr/bin/python

__author__      = "Sander Granneman"
__copyright__   = "Copyright 2019"
__version__     = "0.1.3"
__credits__     = ["Sander Granneman"]
__maintainer__  = "Sander Granneman"
__email__       = "sgrannem@staffmail.ed.ac.uk"
__status__      = "Production"

##################################################################################
#
#   PairedReads
#
#
#   Copyright (c) Sander Granneman 2019
#   
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#   
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#   
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.
#
##################################################################################

from pyCRAC.Classes.Exceptions import *

class PairedReadCollector():
	""" This class is used to TEMPORARILY store information about paired reads.
	To preserve memory, it will only store information for two reads at a time and then automatically starts removing the first read in the list.
	Useful to extract the necessary information from paired reads for analyses and the discard them."""
	def __init__(self):
		"""Initialises the class attributes"""
		self.read_list   = list()

	def addRead(self,interval):
		"""method to add read information to the read dictionary. Hit repeat = "U" indicates that the read has a single alignment location. "R" indicates the reads has multiple alignment locations\
		 read_types: 'L' = first read or read from a forward sequencing reaction, 'R' = second read or read from a reverse sequencing reaction. Object indicates a novoalign or SAM class object. 'line'\
		 indicates a string representing the entire line from the alignment file."""
		if self.numberOfReadsinList() == 2:
			self.removeRead()
		self.read_list.append({
			"chr_ID"        : interval.chr_ID(),
			"read_type"     : interval.read_type(),
			"read_id"       : interval.seq_ID(),
			"strand"        : interval.strand(),
			"read_start"    : interval.read_start(),
			"read_end"      : interval.read_end(),
			"deletions"     : interval.deletions(chromosome_location=True),
			"substitutions" : interval.substitutions(chromosome_location=True),
			"sequence"      : interval.sequence(),
		})
		return True

	def removeRead(self):
		"""removes the first read dictionary from the read list. Useful if two reads are not 'truly' paired"""
		del self.read_list[0]

	def reset(self):
		"""resets the read list and read dictionary"""
		self.read_list = []

	def strand(self,read="L"):
		"""returns the strand of the Forward (L) or Reverse (R) read for properly paired reads"""
		if self.numberOfReadsinList() == 2:
			if read == "L":
				return self.read_list[0]["strand"]
			elif read == "R":
				return self.read_list[1]["strand"]
			else:
				raise InputError("\n\ncould not determine whether you want the strand from the first ('L') or second ('R') read.n\n")
		else:
			return None

	def read_ID(self,read="L"):
		"""returns the sequence name or header of the Forward (L) or Reverse (R) read for properly paired reads"""
		if self.numberOfReadsinList() == 2:
			if read == "L":
				return self.read_list[0]["read_id"]
			elif read == "R":
				return self.read_list[1]["read_id"]
			elif read == "both":
				return (self.read_list[0]["read_id"],self.read_list[1]["read_id"])
			else:
				raise InputError("\n\ncould not determine whether you want the read ID from the first ('L') or second ('R') read\n\n")
		else:
			return None

	def read_type(self,read="L"):
		"""returns the read header of the Forward (L) or Reverse (R) read for properly paired reads"""
		if self.numberOfReadsinList() == 2:
			if read == "L":
				return self.read_list[0]["read_type"]
			elif read == "R":
				return self.read_list[1]["read_type"]
			elif read == "both":
				return (self.read_list[0]["read_type"],self.read_list[1]["read_type"])
			else:
				raise InputError("\n\ncould not determine whether you want the read ID from the first ('L') or second ('R') read\n\n")
		else:
			return None

	def chr_ID(self,read="L"):
		"""returns the read header of the Forward (L) or Reverse (R) read for properly paired reads"""
		if self.numberOfReadsinList() == 2:
			if read == "L":
				return self.read_list[0]["chr_ID"]
			elif read == "R":
				return self.read_list[1]["chr_ID"]
			elif read == "both":
				return (self.read_list[0]["chr_ID"],self.read_list[1]["chr_ID"])
			else:
				raise InputError("\n\ncould not determine whether you want the reference sequence ID from the first ('L') or second ('R') read.\n\n")
		else:
			return None

	def coordinates(self,read="both"):      # read can be "L", "R" or "both". The latter returns min and max coordinates from both reads
		"""returns a tuple with start and end position of a cDNA using paired read coordinates min and max values"""
		if self.numberOfReadsinList() == 2:
			if read == "both":
				coordinates = [self.read_list[0]["read_start"],self.read_list[0]["read_end"],self.read_list[1]["read_start"],self.read_list[1]["read_end"]]
				#print coordinates
				return (min(coordinates),max(coordinates))
			elif read == "L":
				return [self.read_list[0]["read_start"],self.read_list[0]["read_end"]]
			elif read == "R":
				return [self.read_list[1]["read_start"],self.read_list[1]["read_end"]]
			else:
				raise InputError("\n\ncould not determine whether you want the mapping coordinates for the first ('L') or second ('R') read.n\n")
		else:
			return None

	def cDNALength(self):
		"""returns the length of a cDNA using paired read coordinates"""
		positions = self.coordinates()
		if positions:
			start = positions[0]
			end   = positions[1]
			return end-start-1

	def numberOfReadsinList(self):
		"""returns the number of reads present in the self.read_list list"""
		return len(self.read_list)

	def ispaired(self):
		"""checks whether the two reads in the read list are proper pairs and whether the file has the correct order of reads"""
		if self.numberOfReadsinList() == 2:
			if self.read_list[0]["read_type"] == "L" and self.read_list[1]["read_type"] == "R" and self.chr_ID(read="L") == self.chr_ID(read="R"):
				return True
			else:
				return False
		else:
			return False

	def deletions(self,read="both"):                # read can be "L", "R" or "both". The latter returns deletions from all reads
		"""returns a list of all the deletions in the paired reads. Can be genomic positions or read positions"""
		if self.numberOfReadsinList() == 2:
			if read == "both":
				return self.read_list[0]["deletions"] + self.read_list[1]["deletions"]
			elif read == "L":
				return self.read_list[0]["deletions"]
			elif read == "R":
				return self.read_list[1]["deletions"]
			else:
				raise InputError("\n\ncould not determine from which read you want deletions (L,R or both)\n")
		else:
				return None

	def substitutions(self,read="both"):            # read can be "L", "R" or "both". The latter returns substitutions from all reads
		"""returns a list of all the substitutions in the paired reads. Can be genomic positions or read positions"""
		if self.numberOfReadsinList() == 2:
			if read == "both":
				return self.read_list[0]["substitutions"] + self.read_list[1]["substitutions"]
			elif read == "L":
				return self.read_list[0]["substitutions"]
			elif read == "R":
				return self.read_list[1]["substitutions"]
			else:
				raise InputError("\n\ncould not determine from which read you want substitutions (L,R or both)\n" )
		else:
			return None

	def sequence(self,read="both"):                 # read can be "L", "R" or "both". The latter returns sequences from all reads in a list
		"""returns sequences from paired reads"""
		if self.numberOfReadsinList() == 2:
			if read == "both":
				return [self.read_list[0]["sequence"], self.read_list[1]["sequence"]]
			elif read == "L":
				return self.read_list[0]["sequence"]
			elif read == "R":
				return self.read_list[1]["sequence"]
			else:
				raise InputError("\n\ncould not determine from which read you want the sequence (L,R or both)\n")
		else:
			return None

	def mutations(self):
		"""returns a list of all the mutations in the paired reads."""
		return self.deletions(read="both") + self.substitutions(read="both")


#!/usr/bin/python

__author__      = "Sander Granneman"
__copyright__   = "Copyright 2019"
__version__     = "0.0.4"
__credits__     = ["Sander Granneman"]
__maintainer__  = "Sander Granneman"
__email__       = "sgrannem@staffmail.ed.ac.uk"
__status__      = "Production"

##################################################################################
#
#	Aligner
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import re
import bisect
import numpy as np
from collections import defaultdict
from pyCRAC.Classes.Exceptions import *
from pyCRAC.Parsers import ParseAlignments
from pyCRAC.Methods import strand_converter,getfilename,reverse_complement


def highlightMutations(sequence,substitutions,deletions):
	""" Corrects the sequence based on the substitutions and deletions information using info from GTF file data. 
	The substitutions and deletions should be 0-based positions of the read and insertions should have been removed"""
	nucleotides = list(sequence)
	if deletions:
		for position in deletions:
			nucleotides.insert(position,"-")
	if substitutions:
		for position in substitutions:
			nucleotides[position] = nucleotides[position].lower()
	return ''.join(nucleotides)

def seqStart(string):
	"""determines the position of the first nucleotide in a string"""
	a = re.search(r'[atgcnATGCN]',string)
	if a.start():
		return a.start()
	else:
		return 0

def sortByHeadGaps(list_input):
	"""sorts the sequences with head and tail gaps by the number of head gaps"""
	a = [(seqStart(keys.split('\n')[1]),keys) for keys in list_input]
	a.sort()
	return [i[1] for i in a]

def getHeadAndTailGaps(itercoordinates,start,end,gene_strand):
	"""calculates the number of head and tail gaps needed for proper alignment"""
	head = str()
	tail = str()
	index_up = bisect.bisect_left(itercoordinates,start)
	index_down = bisect.bisect_right(itercoordinates,end-1) 
	if gene_strand == "+":
		head = "-" * index_up
		tail = "-" * (len(itercoordinates) - index_down)
	elif gene_strand == "-":
		head = "-" * (len(itercoordinates) - index_down)
		tail = "-" * index_up
	else:
		raise InputError("\n\ngetHeadAndTailGaps method did not receive any strand information.\nPlease double check your input.\n")
	return (head,tail)
	
class ReadAligner():
	def __init__(self,data_file,gtf,genes,file_type="novo",sequence="genomic",blocks=False,base_quality=-50,ignorestrand=False,debug=False):
		""" initializes the class attributes. The ReadAligner class requires at least the path to the data file, a GTF2 object with genes and chromosomes stored.
		If the file type of the data file is not a .novo file then the file_type flag needs to be set to sam or gtf. """
		self.sense_seqs			= defaultdict(list)
		self.anti_sense_seqs	= defaultdict(list)
		self.file_name			= data_file
		self.file_list			= list()
		self.file_type			= file_type.lower()
		self.genes				= genes
		self.data				= None
		self.__blocksremoved	= blocks
		self.__seq_line			= defaultdict(str)
		self.__sequence			= sequence
		self.__gene_sequence	= defaultdict(str)
		self.__itercoordinates	= defaultdict(list)
		self.__alnfile			= None	
		self.__ignorestrand		= ignorestrand
		self.__gtf				= gtf
		if debug: self.file_list.append(debug)
		self.__supported_filetypes = ["novo","sam","gtf"] # index = [0,1,2]; [:2] = novo and sam, [2] = gtf .
		self.__determineFileType(data_file,file_type,debug)

		for gene in self.genes:						   
			if self.__sequence == "coding":
				self.__gene_sequence[gene] = gtf.codingSequence(gene)
			elif self.__sequence == "genomic":
				self.__gene_sequence[gene] = gtf.genomicSequence(gene)
			self.__itercoordinates[gene] = gtf.geneIterCoordinates(gene,coordinates=sequence,output="numpy")
			self.__seq_line[gene] = ">%s\n%s" % (gene,self.__gene_sequence[gene])
					
	def __determineFileType(self,data_file,file_type,debug):
		"""Checks the file type of the data file"""
		if self.file_type in self.__supported_filetypes[:2]:
			self.data = ParseAlignments.ParseAlignmentFile(data_file,file_type,debug=debug)
			self.data.gtfdatatype = None
		elif self.file_type == self.__supported_filetypes[2]:
			self.data = ParseAlignments.ParseCountersOutput(data_file)
		else:
			raise FileTypeError("\n\nThe file type % is not supported by ReadAligner, please choose from the following options: %s\n" % (file_type, ",".join(self.__supported_filetypes))) 
			
#------------------------------------	aligning reads	------------------------------------- 

	def alignReads(self,length=1000,align_quality=0,align_score=0,MAX=-1,limit=5000,unique=False,muts_filter=None):
		"""This method aligns reads from a Novoalign output file (.novo extension) to reference sequences, supplied in a list file
		input required: genes.list file and novoalign file. Options: self,length=1000,flank_seq=0,align_quality=0,align_score=0,correct_sequence=True,MAX=500,sequence="genomic",insertions=False"""
		head_gaps = str()
		line_out  = str()
		tail_gaps = str()
		counter	  = defaultdict(int)
		  
#------------------------------------ if the input file is a novo or sam file -----------------------------------------------	  

		if self.file_type in self.__supported_filetypes[:2]:
			while self.data.getRead(align_quality=align_quality,align_score=align_score,maximum=MAX,blocks=self.__blocksremoved,length=length,unique=unique,force_single_end=True,muts_filter=muts_filter):
				if not self.genes: break
				if muts_filter and not self.data.hasmutation: continue
				for gene in self.genes:
					if counter[gene] == limit: self.genes.remove(gene)
					if self.__gtf.chromosome(gene) == self.data.chromosome:
						overlap = list(self.__itercoordinates[gene][np.logical_and(self.__itercoordinates[gene] >= self.data.read_start,self.__itercoordinates[gene] < self.data.read_end)])
						if len(overlap) == len(self.data.read_seq):
							counter[gene] += 1
							(head_gaps,tail_gaps) = getHeadAndTailGaps(self.__itercoordinates[gene],self.data.read_start,self.data.read_end,self.__gtf.strand(gene))
							if strand_converter(self.data.strand) == self.__gtf.strand(gene):
								dashed_seq = head_gaps+self.data.read_seq+tail_gaps
								line_out = ">%s\n%s" % (self.data.read_seq_ID,dashed_seq)
								self.sense_seqs[gene].append(line_out)		  
									
							elif self.__ignorestrand:
								dashed_seq = head_gaps+reverse_complement(self.data.read_seq)+tail_gaps
								line_out = ">%s\n%s" % (self.data.read_seq_ID,dashed_seq)
								self.sense_seqs[gene].append(line_out)	 
		
							elif strand_converter(self.data.strand) != self.__gtf.strand(gene):
								dashed_seq = head_gaps+reverse_complement(self.data.read_seq)+tail_gaps
								line_out = ">%s\n%s" % (self.data.read_seq_ID,dashed_seq)
								self.anti_sense_seqs[gene].append(line_out)
															
#------------------------------------ if the input file is a pyReadCounter/pyMotif GTF file ----------------------------------- 

		elif self.file_type == self.__supported_filetypes[2]:
			while self.data.readLineByLine():
				if not self.genes: break
				genesinboth = set(self.genes) & set(self.data.genes)
				if genesinboth:
					for gene in genesinboth:
						if counter[gene] == limit: 
							self.genes.remove(gene)
						if self.__gtf.chromosome(gene) == self.data.chromosome:
							sequence = self.__gtf.sequence(self.data.chromosome,strand_converter(self.data.strand),self.data.read_start,self.data.read_end)
							substitutions = [i - self.data.read_start for i in self.data.substitutions] 
							deletions	  = [i - self.data.read_start for i in self.data.deletions]
							self.data.read_seq	  = highlightMutations(sequence,substitutions,deletions)
							self.data.read_seq_ID = "%s_%s_%s" % (gene,self.data.read_start,self.data.read_end)
							overlap = list(self.__itercoordinates[gene][np.logical_and(self.__itercoordinates[gene] >= self.data.read_start,self.__itercoordinates[gene] < self.data.read_end)])
							if len(overlap) == len(self.data.read_seq):
								counter[gene] += 1
								(head_gaps,tail_gaps) = getHeadAndTailGaps(self.__itercoordinates[gene],self.data.read_start,self.data.read_end,self.__gtf.strand(gene))
								if strand_converter(self.data.strand) == self.__gtf.strand(gene):
									dashed_seq = head_gaps+self.data.read_seq+tail_gaps
									line_out = ">%s\n%s" % (self.data.read_seq_ID,dashed_seq)
									self.sense_seqs[gene].append(line_out)		  
										
								elif self.__ignorestrand:
									dashed_seq = head_gaps+reverse_complement(self.data.read_seq)+tail_gaps
									line_out = ">%s\n%s" % (self.data.read_seq_ID,dashed_seq)
									self.sense_seqs[gene].append(line_out)	 
			
								elif strand_converter(self.data.strand) != self.__gtf.strand(gene):
									dashed_seq = head_gaps+reverse_complement(self.data.read_seq)+tail_gaps
									line_out = ">%s\n%s" % (self.data.read_seq_ID,dashed_seq)
									self.anti_sense_seqs[gene].append(line_out)										 
														  

	#------------------------------------ printing the alignments -------------------------------------

	def __createFileHandles(self,gene,file_type,file_extension="fasta"):
		""" does what it says """
		filename = str()								 
		filename = "%s-%s_%s_%s_%s.%s" % (file_type,self.__insert,gene,self.__sequence,getfilename(self.file_name),file_extension)
		self.file_list.append(filename)
		return filename
	
	def printFastaWithAlignments(self,out_file=False):
		self.__insert = "reads"
		if self.data.gtfdatatype:
			if self.data.gtfdatatype == "blocks":
				self.__insert = "cDNAs"
			elif self.data.gtfdatatype == "clusters":
				self.__insert = "clusters"
			elif self.data.gtfdatatype == "peaks":
				self.__insert = "peaks"
		else:
			if self.__blocksremoved:
				self.__insert = "cDNAs"
		if not self.sense_seqs and not self.anti_sense_seqs:
			raise NoResultsError("\n\nCould not find any reads that matched your criteria.\nPlease double check your settings.\n")
		file_out = None
		if out_file:
			if out_file == "default":
				file_name = "%s_%s_%s.fasta" % (self.__insert,getfilename(self.file_name),self.__sequence)
			else:
				file_name = out_file
			file_out = open(file_name,"w")
			self.file_list.append(file_name)
			for gene in self.__gene_sequence:
				if self.sense_seqs[gene]:
					sorted_tuple_list = sortByHeadGaps(self.sense_seqs[gene])
					file_out.write("### %s sense %s ###\n" % (gene,self.__insert))
					file_out.write("%s\n" % self.__seq_line[gene])
					for lines in sorted_tuple_list:
							file_out.write("%s\n" % lines)
					file_out.write("\n\n")
				if self.anti_sense_seqs[gene]:
					sorted_tuple_list = sortByHeadGaps(self.anti_sense_seqs[gene])
					file_out.write("### %s anti-sense %s ###\n" % (gene,self.__insert))
					file_out.write("%s\n" % self.__seq_line[gene])
					for lines in sorted_tuple_list:
							file_out.write("%s\n" % lines)
					file_out.write("\n\n")	
			file_out.close()
		else:
			for gene in self.__gene_sequence:
				if self.sense_seqs[gene]:
					sorted_tuple_list = sortByHeadGaps(self.sense_seqs[gene])
					file_type = "sense"
					file_name = self.__createFileHandles(gene,file_type,file_extension="fasta")
					file_out  = open(file_name,"w")
					file_out.write("%s\n" % self.__seq_line[gene])
					for lines in sorted_tuple_list:
							file_out.write("%s\n" % lines)
					file_out.close()

				if self.anti_sense_seqs[gene]:
					sorted_tuple_list = sortByHeadGaps(self.anti_sense_seqs[gene])
					file_type = "anti_sense"
					file_name = self.__createFileHandles(gene,file_type,file_extension="fasta")
					file_out  = open(file_name,"w")								 
					file_out.write("%s\n" % self.__seq_line[gene])
					for lines in sorted_tuple_list:
							file_out.write("%s\n" % lines)
					file_out.close()